/******************************************************************
author: qingzhe huang
date: Jul. 31, 2007
1. This module is a combination of both "Master" and "Displayer". Originally it
should be separated, at least logically. However, in order to run in more nodes, I 
have to save displayer to combine it with master nodes as displayer is not a heavy
job.
2. The most difficult part is "tile" which troubles me for many days. Originally I
don't really grasp the meaning of "glReadPixel" which is so powerful that allow programer
to read only "tiles". You see, in both Testers and Renders they are using "tiles" 
to divide screen. However, when sending pixel data, Renders send pixel data according
to "tiles". Then using "tile" as basic unit, you don't have to worry about "sequence".
i.e. all "linear" sequence in both "pixel-visible-info" and "pixel-data" are based on 
"tiles". You just move your pointer and finally write back to color buffer in "tiles".
The big advantage is that in network transmission data buffer needs to be consecutive and
you don't have to make extra copy if you read and write color buffer in "tiles". This 
may not be a problem at all for somebody. But for me at beginning I read whole color 
buffer in one shot and manually divide "tiles" by index which is really painful experience.
3. The so-called knapsack is a greedy for NP-complete problem. I don't want to waste 
too much time on it, so make it a simple greedy.


******************************************************************/


#include "config.h"
#include "displayer.h"
#include "dataStructure.h"
#include <assert.h>
#include <GL/glut.h>

#ifndef  LINUX_VERSION
#include "except.h"
#endif

using namespace std;



char* DisplayerContext::windowTitle="Displayer Node";
extern MPIContext mpiContext;


void DisplayerContext::init()
{
	int i;
	colorBuffer=new ubyte[ColorBufferByteSize];
	CHECK_NEW(colorBuffer);

	colorBufferRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(colorBufferRequests);

	pixelDataBuffer=new ubyte[ColorBufferByteSize];
	CHECK_NEW(pixelDataBuffer);

	visibleTable=new ubyte[ColorBufferSize];
	CHECK_NEW(visibleTable);

	compressedVisibleTable=new ubyte[ColorBufferSize*VisibleTableBufferRate];
	CHECK_NEW(compressedVisibleTable);

	visibleTableRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(visibleTableRequests);
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		visibleTableRequests[i]=MPI_REQUEST_NULL;
	}

	pixelDataRequests=new MPI_Request[RenderCount];
	CHECK_NEW(pixelDataRequests);
	for (i=0; i<RenderCount; i++)
	{
		pixelDataRequests[i]=MPI_REQUEST_NULL;
	}

	glMatrixMode(GL_PROJECTION);
	glOrtho(0, ScreenSize, 0, ScreenSize, 0, 100);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glDrawBuffer(GL_BACK);
}



void DisplayerContext::recvVisiblePixelInfo()
{
	for (int i=0; i<DepthBufferSnapShotNumber; i++)
	{
		visibleTableRequests[i]=MPI_REQUEST_NULL;
#ifndef USING_COMPRESSION
		SAFE_CALL(MPI_Irecv(visibleTable+i*DepthBufferAtomSize, DepthBufferAtomSize, MPI_UNSIGNED_CHAR,
			MPI_ANY_SOURCE, i, MPI_COMM_WORLD, visibleTableRequests+i));
#else
		SAFE_CALL(MPI_Irecv(compressedVisibleTable+i*DepthBufferAtomSize*VisibleTableBufferRate, DepthBufferAtomSize*VisibleTableBufferRate,
			 MPI_UNSIGNED_CHAR, MPI_ANY_SOURCE, i, MPI_COMM_WORLD, visibleTableRequests+i));
#endif		
	}
	DEBUG_INFO("displayer receives visible pixel info \n");
	//	runningStruct[RenderHostCount].arrayIndex, renderContext.frameNumber);
}

//let us use wait_any later for optimization
void DisplayerContext::testVisiblePixelInfoRecving()
{
	int i;
#ifdef 	USING_COMPRESSION
	MPI_Status status[DepthBufferSnapShotNumber];
	int length;
	
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, visibleTableRequests, status));
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		SAFE_CALL(MPI_Get_count(status+i, MPI_UNSIGNED_CHAR, &length));
		length=decompressBuffer(compressedVisibleTable+i*DepthBufferAtomSize*VisibleTableBufferRate, length, 
			visibleTable+i*DepthBufferAtomSize, DepthBufferAtomSize);
		assert(length==DepthBufferAtomSize);
	}
#else
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, visibleTableRequests, MPI_STATUSES_IGNORE));
#endif
	DEBUG_INFO("displayer finishes visible pixel info receiving\n");
}

void DisplayerContext::displayerPrepareRecvPixelData()
{
	int i;
	for (i=0; i<RenderCount+1; i++)
	{
		renderPixelCount[i]=0;
	}
	//counting how many pixel from each render, this is purely to save memory
	//so that in total we only need one color buffer size of buffer to hold all pixel data
	//from all renders. 
	for (i=0; i<ColorBufferSize; i++)
	{
		renderPixelCount[visibleTable[i]]++;
	}
}

void DisplayerContext::recvColorBuffer()
{
	int i, offset=0;	
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		//here we use single size of color buffer to hold all pixel data from all renders
		SAFE_CALL(MPI_Irecv(colorBuffer+ColorBufferAtomSize*i, ColorBufferAtomSize, MPI_UNSIGNED_CHAR, 
				MPI_ANY_SOURCE, i, MPI_COMM_WORLD, colorBufferRequests+i));
		DEBUG_INFO("displayer begin recv pixel data\n");
	}
	DEBUG_INFO("displayer post requests to receive pixel data\n");
}

void DisplayerContext::testColorBufferRecving()
{
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, colorBufferRequests, MPI_STATUSES_IGNORE));
}



void DisplayerContext::recvPixelData()
{
	int i, offset=0;	
	for (i=0; i<RenderCount; i++)
	{
		//here we use single size of color buffer to hold all pixel data from all renders
		renderPixelPtr[i]=pixelDataBuffer+offset;
		if (renderPixelCount[i]>0)
		{
			SAFE_CALL(MPI_Irecv(renderPixelPtr[i], renderPixelCount[i]*4, MPI_UNSIGNED_CHAR, 
				RenderHostStartIndex+i, PIXEL_DATA, MPI_COMM_WORLD, pixelDataRequests+i));
			DEBUG_INFO("displayer begin recv pixel data\n");
		}
		offset+=renderPixelCount[i]*4;		
		//printf("displayer expects to receives pixel data of %d from render#%d \n", renderPixelCount[i], i);
	}
	DEBUG_INFO("displayer post requests to receive pixel data\n");
}



//first let's do the simplest
void DisplayerContext::testPixelDataRecving()
{
	int i;
	SAFE_CALL(MPI_Waitall(RenderCount, pixelDataRequests, MPI_STATUSES_IGNORE));
	for (i=0; i<ColorBufferSize; i++)
	{
		if (visibleTable[i]<RenderCount)
		{
			//visible
			memcpy(colorBuffer+i*4, renderPixelPtr[visibleTable[i]], 4);
			renderPixelPtr[visibleTable[i]]+=4;
		}
		else
		{
			//set unvisible background to white
			memset(colorBuffer+i*4, 255, 4);
		}
	}
	//printf("displayer finishes receiving pixel data of frame#%d\n", mpiContext.renderContext->frameNumber);
	DEBUG_INFO("Displayer finishes pixel data recving\n");
}


int myFrameCount=0;

#ifdef	 USING_TESTER_RENDER
void DisplayerContext::displayCallback()
{
	DEBUG_INFO("displayer begins displaycallback......\n");
	//PROFILE_PRINT();
	PROFILE_BEGIN(ALL);


#ifdef	USING_SYNCHRONIZATION
	PROFILE_BEGIN(BMTRD);
	mpiContext.recvRenderContext();
	mpiContext.testRenderContextRecving();
	PROFILE_END(BMTRD);
	if (mpiContext.startMPI)
	{
		PROFILE_BEGIN(BRD);
		recvColorBuffer();
		testColorBufferRecving();
		PROFILE_END(BRD);
		PROFILE_BEGIN(CD);
		outputColorBuffer();
		PROFILE_END(CD);
		PROFILE_END(ALL);
	}
#else
	//printf("displayer frameCount=%d\n", myFrameCount);
	if (myFrameCount==QUIT_FRAME_NUMBER-START_FRAME_NUMBER)
	{
		mpiContext.quit();
	}
	myFrameCount++;
	
	PROFILE_BEGIN(BRD);
	recvColorBuffer();
	testColorBufferRecving();
	PROFILE_END(BRD);
	PROFILE_BEGIN(CD);
	outputColorBuffer();
	PROFILE_END(CD);
	PROFILE_END(ALL);
	DEBUG_INFO("displayer finishes displaycallback......\n");
#endif

	DEBUG_INFO("displayer finishes displaycallback......\n");

}
#else//
void DisplayerContext::displayCallback()
{
	DEBUG_INFO("displayer begins displaycallback......\n");
	//PROFILE_PRINT();
	PROFILE_BEGIN(ALL);


#ifdef	USING_SYNCHRONIZATION
	PROFILE_BEGIN(BMTRD);
	mpiContext.recvRenderContext();
	mpiContext.testRenderContextRecving();
	PROFILE_END(BMTRD);
	if (mpiContext.startMPI)
	{
		PROFILE_BEGIN(BTRD);
		recvVisiblePixelInfo();
		testVisiblePixelInfoRecving();
		displayerPrepareRecvPixelData();
		PROFILE_END(BTRD);
	
		PROFILE_BEGIN(BRD);
		recvPixelData();
		testPixelDataRecving();
		PROFILE_END(BRD);
		PROFILE_BEGIN(CD);
		outputColorBuffer();
		PROFILE_END(CD);
		PROFILE_END(ALL);
	}
#else
	//printf("displayer frameCount=%d\n", myFrameCount);
	if (myFrameCount==QUIT_FRAME_NUMBER-START_FRAME_NUMBER)
	{
		mpiContext.quit();
	}
	myFrameCount++;
	PROFILE_BEGIN(BTRD);
	recvVisiblePixelInfo();
	testVisiblePixelInfoRecving();
	displayerPrepareRecvPixelData();
	PROFILE_END(BTRD);
	
	PROFILE_BEGIN(BRD);
	recvPixelData();
	testPixelDataRecving();
	PROFILE_END(BRD);
	PROFILE_BEGIN(CD);
	outputColorBuffer();
	PROFILE_END(CD);
	PROFILE_END(ALL);
	DEBUG_INFO("displayer finishes displaycallback......\n");
#endif

	DEBUG_INFO("displayer finishes displaycallback......\n");

}
#endif

void DisplayerContext::outputColorBuffer()
{
	int i, j;
	//save context and setup to write color buffer
	//glOrtho(-ScreenSize/2, ScreenSize/2, -ScreenSize/2, ScreenSize/2, DefaultNearPlane, 100);
	//glRasterPos2f(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth);
	glRasterPos2f(0, 0);
	for (i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (j=0; j<DepthBufferSnapShotWidth; j++)
		{			
			//if (j%2==0)
			glRasterPos2f(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth);
			glDrawPixels(DepthBufferAtomWidth, DepthBufferAtomWidth, 
				GL_RGBA, GL_UNSIGNED_BYTE, colorBuffer+(i*DepthBufferSnapShotWidth+j)*DepthBufferAtomSize*4);
		}
	}
	DEBUG_INFO("Displayer finishes output color buffer\n");
}


